from typing import Tuple

from easypage.conn import Conn


class Browser:
    """
        Browser

        https://chromedevtools.github.io/devtools-protocol/tot/Browser
    """

    def __init__(self, conn: Conn):
        self.__conn = conn

    def version(
            self,
            raise_err: bool = False,
            need_callback: bool = True,
    ) -> Tuple[bool, dict]:
        """
        获取浏览器的信息

        https://chromedevtools.github.io/devtools-protocol/tot/Browser/#method-getVersion

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="Browser.getVersion",
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def close(
            self,
            raise_err: bool = False,
            need_callback: bool = False,
    ) -> Tuple[bool, dict]:
        """
        退出浏览器

        https://chromedevtools.github.io/devtools-protocol/tot/Browser/#method-close

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="Browser.close",
            raise_err=raise_err,
            need_callback=need_callback,
        )



from typing import Tuple
from easypage.conn import Conn


class Target:
    """
        Target

        https://chromedevtools.github.io/devtools-protocol/tot/Target
    """

    def set_discover_targets(
            self,
            discover: bool = True,
            raise_err: bool = False,
            need_callback: bool = False,
            **kwargs
    ) -> Tuple[bool, dict]:
        """
        控制是否发现可用目标并通过 targetCreated/targetInfoChanged/targetDestroyed 事件进行通知

        https://chromedevtools.github.io/devtools-protocol/tot/Target/#method-setDiscoverTargets

        :param discover: 是否发现可用目标
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        send_data = {
            "discover": discover
        }

        send_data.update(kwargs)

        return self.__conn.send(
            method="Target.setDiscoverTargets",
            params=send_data,
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def __init__(self, conn: Conn):
        self.__conn = conn

    def get_targets(
            self,
            raise_err: bool = False,
            need_callback: bool = True,
    ) -> Tuple[bool, dict]:
        """
        获取所有的 target

        https://chromedevtools.github.io/devtools-protocol/tot/Target/#method-getTargets

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="Target.getTargets",
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def get_browser_contexts(
            self,
            raise_err: bool = False,
            need_callback: bool = True,
            **kwargs
    ) -> Tuple[bool, dict]:
        """
        获取所有的浏览器上下文

        https://chromedevtools.github.io/devtools-protocol/tot/Target/#method-getBrowserContexts

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="Target.getBrowserContexts",
            params=kwargs,
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def create_browser_context(
            self,
            proxy: str = None,
            proxy_pass_list: list = None,
            raise_err: bool = False,
            need_callback: bool = True,
            **kwargs
    ) -> Tuple[bool, dict]:
        """
        新建上下文

        学习链接：https://chromedevtools.github.io/devtools-protocol/tot/Target/#method-createBrowserContext

        :param proxy: 代理 如：http://127.0.0.1:7890
        :param proxy_pass_list: 代理绕过列表，类似于传递给 --proxy-bypass-list 的列表（表示某些地址不走代理）如：["www.baidu.com"]
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :param kwargs:
        :return:
        """
        if proxy and not proxy.startswith("http"):
            proxy = "http://" + proxy

        send_data = {
            "disposeOnDetach": True,
            "proxyServer": proxy or "",
            "proxyBypassList": ",".join(proxy_pass_list) if proxy_pass_list else ""
        }

        send_data.update(kwargs)

        return self.__conn.send(
            method='Target.createBrowserContext',
            params=send_data,
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def create_target(
            self,
            url: str = None,
            browser_context_id: str = None,
            raise_err: bool = False,
            need_callback: bool = True,
            **kwargs
    ) -> Tuple[bool, dict]:
        """
        创建新的页面

        https://chromedevtools.github.io/devtools-protocol/tot/Target/#method-createTarget

        :param url:
        :param browser_context_id:
        :param kwargs:
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        send_data = {
            "url": url or "about:blank",
        }

        if browser_context_id:
            send_data["browserContextId"] = browser_context_id

        send_data.update(kwargs)

        return self.__conn.send(
            method='Target.createTarget',
            params=send_data,
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def close_target(
            self,
            target_id: str,
            raise_err: bool = False,
            need_callback: bool = False,
    ) -> Tuple[bool, dict]:
        """
        关闭 target

        https://chromedevtools.github.io/devtools-protocol/tot/Target/#method-closeTarget

        :param target_id:
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="Target.closeTarget",
            params={
                "targetId": target_id
            },
            raise_err=raise_err,
            need_callback=need_callback,
        )



import socket
from pathlib import Path
from platform import system
from typing import List, Union


def find_browser_path() -> str:
    """
    寻找浏览器的路径

    :return:
    """
    # 通过 which 获取
    from shutil import which

    which_list = [
        'chrome',
        'chromium',
        'google-chrome',
        'google-chrome-stable',
        'google-chrome-unstable',
        'google-chrome-beta',
    ]
    for wl in which_list:
        path = which(wl)
        if path:
            return path

    # 默认路径获取
    sys_type = system().lower()
    if sys_type in ('macos', 'darwin'):
        p = '/Applications/Google Chrome.app/Contents/MacOS/Google Chrome'
        return p if Path(p).exists() else None

    elif sys_type == 'linux':
        paths = ('/usr/bin/google-chrome', '/opt/google/chrome/google-chrome',
                 '/user/lib/chromium-browser/chromium-browser')
        for p in paths:
            if Path(p).exists():
                return p
        return

    elif sys_type == 'windows':
        # 通过注册表获取
        from winreg import OpenKey, EnumValue, CloseKey, HKEY_CURRENT_USER, HKEY_LOCAL_MACHINE, KEY_READ

        txt = r'SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\chrome.exe'
        for find_key in [HKEY_CURRENT_USER, HKEY_LOCAL_MACHINE]:
            try:
                key = OpenKey(find_key, txt, reserved=0, access=KEY_READ)
                k = EnumValue(key, 0)
                CloseKey(key)
                if k[1]:
                    return k[1]
            except (FileNotFoundError, OSError):
                pass
    else:
        raise Exception(f"不受支持的系统：{sys_type}")


def find_free_port() -> int:
    """
    寻找一个空闲端口

    :return:
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    # 绑定到本地地址和端口，让系统自动分配端口
    sock.bind(('127.0.0.1', 0))

    # 获取分配的端口
    port = sock.getsockname()[1]

    sock.close()

    return port


def port_in_use(port, ip: str = '127.0.0.1') -> bool:
    """
    判断指定端口是否被占用

    :param port:
    :param ip:
    :return:
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(0.1)
    result = sock.connect_ex((ip, port))
    sock.close()

    return result == 0


def parse_driver_id(data: Union[List[dict], dict]) -> List[str]:
    """
    返回 id 列表

    :param data:
    :return:
    """
    if not isinstance(data, list):
        data = [data]

    result = []
    for d in data:
        if d.get("webSocketDebuggerUrl"):
            result.append(d['webSocketDebuggerUrl'].split('/')[-1])

    return result


def parse_box_center_coordinate(data: list) -> tuple:
    """
    根据给定的目标，计算中心点坐标

    :param data: [844, 222.703125, 952, 222.703125, 952, 266.703125, 844, 266.703125]
    :return:
    """
    # 提取四个顶点的坐标
    x1, y1, x2, y2, x3, y3, x4, y4 = data

    # 计算矩形的左上角和右下角坐标
    left = min(x1, x4)
    right = max(x2, x3)
    top = min(y1, y2)
    bottom = max(y3, y4)

    # 计算矩形中心点的坐标
    center_x = (left + right) / 2
    center_y = (top + bottom) / 2

    return center_x, center_y


if __name__ == "__main__":
    free_port = find_free_port()
    print("Free port:", free_port)

    port_to_check = 9222
    if port_in_use(port_to_check):
        print(f"Port {port_to_check} is in use.")
    else:
        print(f"Port {port_to_check} is free.")



from typing import Tuple
from easypage.conn import Conn


class Input:
    """
        Input

        https://chromedevtools.github.io/devtools-protocol/tot/Input
    """

    def __init__(self, conn: Conn):
        self.__conn = conn

    def dispatch_key_event(
            self,
            event_type: str,
            raise_err: bool = False,
            need_callback: bool = False,
            **kwargs
    ) -> Tuple[bool, dict]:
        """
        执行键盘的命令

        学习链接：https://chromedevtools.github.io/devtools-protocol/tot/Input/#method-dispatchKeyEvent

        :param event_type: 如：keyDown, keyUp, rawKeyDown, char
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        send_data = {
            "type": event_type,
        }
        send_data.update(kwargs)

        if "key" in send_data:
            send_data['key'] = "Key " + send_data['key'].upper()

        return self.__conn.send(
            method="Input.dispatchKeyEvent",
            params=send_data,
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def dispatch_mouse_event(
            self,
            event_type: str,
            x: int,
            y: int,
            button: str = "none",
            raise_err: bool = False,
            need_callback: bool = False,
            **kwargs
    ) -> Tuple[bool, dict]:
        """
        执行鼠标的命令

        学习链接：https://chromedevtools.github.io/devtools-protocol/tot/Input/#method-dispatchMouseEvent

        :param event_type: 如：mousePressed, mouseReleased, mouseMoved, mouseWheel
        :param x: 点击的坐标
        :param y: 点击的坐标
        :param button: 代表鼠标的按键，如：none（不触发按键，代表光标）, left（左击）, middle（滚轮）, right（右击）, back（后退）, forward（前进）
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        send_data = {
            "type": event_type,
            "x": x,
            "y": y,
            "button": button,
        }
        send_data.update(kwargs)

        return self.__conn.send(
            method="Input.dispatchMouseEvent",
            params=send_data,
            raise_err=raise_err,
            need_callback=need_callback,
        )



from typing import Tuple
from easypage.conn import Conn


class Security:
    """
        Storage

        https://chromedevtools.github.io/devtools-protocol/tot/Security
    """

    def __init__(self, conn: Conn):
        self.__conn = conn

    def enable(
            self,
            raise_err: bool = False,
            need_callback: bool = False,
    ) -> Tuple[bool, dict]:
        """
        启用安全

        https://chromedevtools.github.io/devtools-protocol/tot/Security/#method-enable

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="Security.enable",
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def disable(
            self,
            raise_err: bool = False,
            need_callback: bool = False,
    ) -> Tuple[bool, dict]:
        """
        禁用安全

        https://chromedevtools.github.io/devtools-protocol/tot/Security/#method-disable

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="Security.disable",
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def set_ignore_certificate_errors(
            self,
            ignore: bool = True,
            raise_err: bool = False,
            need_callback: bool = False,
    ) -> Tuple[bool, dict]:
        """
        忽略证书错误

        https://chromedevtools.github.io/devtools-protocol/tot/Security/#method-setIgnoreCertificateErrors

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :param ignore:
        :return:
        """
        return self.__conn.send(
            method="Security.setIgnoreCertificateErrors",
            params={
                "ignore": ignore
            },
            raise_err=raise_err,
            need_callback=need_callback,
        )



"""
    一些浏览器上面的操作
"""
import time
import requests
from easypage import errors
from subprocess import Popen
from easypage import settings
from easypage.logger import logger
from easypage.utils import browser_utils
from easypage.options import BrowserOptions


def browser(address: str) -> dict:
    """
    获取浏览器信息

    :param address:
    :return:
    """
    resp = requests.get(
        url=f'http://{address}/json/version',
        headers={
            'Connection': 'close'
        },
        timeout=10,
    )

    if resp.status_code != 200:
        raise Exception("浏览器连接失败！")

    return resp.json()


def pages(address: str) -> list:
    """
    获取所有的页面信息

    :param address:
    :return:
    """
    resp = requests.get(
        url=f'http://{address}/json',
        headers={
            'Connection': 'close'
        },
        timeout=10,
    )

    return list(filter(
        lambda i: i['type'] in ('page') and not i['url'].startswith('devtools://'),
        resp.json()
    ))


def new_browser(opts: BrowserOptions, check_port_use: bool = True):
    """
    新建浏览器

    :param opts:
    :param check_port_use: 检测端口使用状态
    :return:
    """
    # 获取路径
    browser_path = opts.browser_path or browser_utils.find_browser_path()
    if not browser_path:
        raise FileNotFoundError("浏览器执行路径未找到！")
    logger.debug("本次浏览器连接路径：{}", browser_path)

    if settings.SYSTEM_TYPE == "linux":
        if not opts.exists("--no-sandbox"):
            opts.add("--no-sandbox")
        if not opts.headless:
            logger.warning("当前系统为 linux 但是未设置 --headless")

    # 判断端口使用状态
    port = opts.port or settings.BROWSER_START_PORT
    if check_port_use:
        if browser_utils.port_in_use(port):
            port = browser_utils.find_free_port()
            opts.set_address(ip=opts.ip or settings.BROWSER_START_IP, port=port)
            logger.warning("当前端口被占用，已启用随机端口：{}", port)

    # 判断用户文件夹（不创建的话，会和已有的混在一起导致异常：）
    if not opts.user_data_dir:
        opts.set_user_data_dir(path=settings.USER_DATA_DIR + str(opts.port), delete_if_exists=True)
        logger.debug("未设置用户文件夹，已使用默认路径：{}", opts.user_data_dir)

    # 加入插件
    if opts.extensions:
        opts.add(f"--load-extension={','.join(opts.extensions)}")

    # headless 判断（隐私模式没加 ua 的话，会导致部分网站通过 ua 检测出来：Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) HeadlessChrome/125.0.0.0 Safari/537.36）
    if opts.headless and not opts.user_agent:
        opts.set_user_agent(user_agent=settings.USER_AGENT)

    # 启动浏览器
    logger.debug("本次浏览器连接端口：{}", opts.port)
    arguments = [str(browser_path), f'--remote-debugging-address={opts.ip}', f'--remote-debugging-port={opts.port}']
    arguments.extend(opts.args)

    try:
        return Popen(arguments, shell=False)
    except:
        logger.error("浏览器启动失败！")
        raise


def wait_connection(opts: BrowserOptions):
    """
    等待连接

    :param opts:
    :return:
    """
    start_time = time.time()

    while time.time() - start_time < opts.timeout:
        if test_connection(opts):
            return

    raise errors.BrowserConnectTimeout(opts.address)


def test_connection(opts: BrowserOptions) -> bool:
    """
    测试连通性

    :param opts:
    :return:
    """
    try:
        resp = requests.get(
            url=f'http://{opts.address}/json/version',
            headers={
                'Connection': 'close'
            },
            timeout=30,
        )
        if resp.status_code == 200:
            return True
    except:
        pass

    return False



"""
    页面 driver
"""
import json
import re
import threading
import time
import base64
import websocket
from pathlib import Path
from parsel import Selector
from bs4 import BeautifulSoup
from easypage.listener import Listener
from easypage.logger import logger
from easypage.driver.driver import Driver
from easypage.options import BrowserOptions
from typing import Tuple, Union, List, Callable
from easypage import definition, errors, operate
from easypage.conn import TYPE_PAGE, DRIVER_TYPE
from easypage.request import Request
from easypage.response import Response
from easypage.utils import other_utils, browser_utils


class PageDriver(Driver):
    def __init__(
            self,
            opts: BrowserOptions,
            page_id: str,
            context_id: str,
            context_driver,
            proxy: definition.Proxy = None
    ):
        """

        :param opts: 配置
        :param page_id: driver_id
        :param context_id: 所属上下文 id
        :param context_driver: 所属上下文 driver
        :param proxy: 代理
        """
        super().__init__(opts, TYPE_PAGE, page_id)
        self.page_id = page_id
        self.context_id = context_id
        self._proxy = proxy
        self._opts = opts
        self._context_driver = context_driver

        # TODO 查找一下是否有可能出现未捕获到请求，但是出现响应的情况
        # 预设参数
        self.__loading: bool = True
        self.__requests_cache = {}  # 可能会有相同请求，request_id 也一样，所以不能移除
        self.__response_cache = {}

        # TODO 考虑根据不同的条件去设置是否启动监听避免额外的资源消耗，可以根据事件在 Listener 进行驱动
        # 启用事件
        self.cdp.dom.enable()
        self.cdp.page.enable()
        self.cdp.security.set_ignore_certificate_errors()
        self.cdp.fetch.enable(handle_auth_requests=True)
        self.cdp.network.enable()
        self.cdp.css.enable()

        # 事件监听
        self.cdp.on(event='Page.lifecycleEvent', callback=self.__on_page_life_cycle_event)
        self.cdp.on(event='Page.loadEventFired', callback=self.__on_page_load_event_fired)
        self.cdp.on(event='Page.frameStartedLoading', callback=self.__on_page_frame_started_loading)
        self.cdp.on(event='Page.frameStoppedLoading', callback=self.__on_page_frame_stopped_loading)
        self.cdp.on(event='Fetch.requestPaused', callback=self.__on_fetch_request_paused)
        self.cdp.on(event='Fetch.authRequired', callback=self.__on_fetch_auth_required)
        self.cdp.on(event='Network.requestWillBeSent', callback=self.__on_network_request_will_be_sent)
        self.cdp.on(event='Network.requestWillBeSentExtraInfo',
                    callback=self.__on_network_request_will_be_sent_extra_info)
        self.cdp.on(event='Network.responseReceived', callback=self.__on_network_response_received)
        self.cdp.on(event='Network.loadingFinished', callback=self.__on_network_loading_finished)

    def get(self, url, **kwargs) -> Tuple[bool, Union[None, str]]:
        """
        访问某个页面

        返回的第一个值是请求命令是否成功
        返回的第二个值是当请求失败的时候发送原因，比如设置的代理访问不通

        :param url:
        :param kwargs:
        :return:
        """
        if "frameId" not in kwargs:
            kwargs["frameId"] = self.page_id
        status, result = self.cdp.page.navigate(url=url, **kwargs)
        if not status:
            return False, None

        elif "errorText" in result:
            logger.warning("url：{} errorText：{}", url, result["errorText"])
            return False, result["errorText"]

        return True, None

    def query_selector(
            self,
            selector: str,
            node_id: int = None,
            raise_err: bool = False
    ) -> Union['Element', None, errors.ElementNotFound]:
        """
        选择元素

        :param selector: 选择器
        :param node_id: 节点 id（基于某个节点查找时需要）
        :param raise_err: 是否报错
        :return:
        """
        status, new_node = self.cdp.dom.query_selector(
            selector=selector,
            node_id=node_id,
            raise_err=raise_err,
        )
        if (status and new_node['nodeId'] == 0) or not new_node or not status:
            if raise_err:
                raise errors.ElementNotFound(selector)
            return None

        return Element(node_id=new_node['nodeId'], page=self)

    def query_selector_all(
            self,
            selector: str,
            node_id: int = None,
            raise_err: bool = False
    ) -> Union[List['Element'], None, errors.ElementNotFound]:
        """
        选择元素

        :param selector: 选择器（支持 css、xpath）
        :param node_id: 节点 id（基于某个节点查找时需要）
        :param raise_err: 是否报错
        :return:
        """
        status, new_nodes = self.cdp.dom.query_selector_all(
            selector=selector,
            node_id=node_id,
            raise_err=raise_err,
        )
        if not status or not new_nodes:
            if raise_err:
                raise errors.ElementNotFound(selector)
            return None

        elements = []
        for nn in new_nodes['nodeIds']:
            elements.append(Element(node_id=nn, page=self))
        return elements

    def eval_js(self, js: str, **kwargs) -> Tuple[bool, any]:
        """
        执行 js 脚本

        :param js: 路径或者 js 字符串
        :return: 可能是字符串、数字等
        """
        status, result = self.cdp.runtime.evaluate(js, **kwargs)
        if not status:
            return status, result

        if 'value' in result['result']:
            return True, result['result']['value']
        else:
            return False, result['result']['description']

    def iframes(self) -> List["Element"]:
        """
        获取所有的 iframe

        :return:
        """
        # operate.pages(self._opts.address)

        elements = []

        status, new_node_ids = self.cdp.dom.query_selector_all(
            selector="iframe",
        )
        if not new_node_ids or not status:
            return elements

        for node_id in new_node_ids['nodeIds']:
            # 获取 frameId
            status, frame_detail = self.cdp.dom.describe_node(nodeId=node_id)
            if not frame_detail or not status:
                continue
            frame_id = frame_detail['node']['frameId']

            # 判断是否同域
            status, frame_tree = self.cdp.page.get_frame_tree()
            if not status:
                continue
            frame_tree = frame_tree["frameTree"]

            """
                小知识点
                同域名的情况下，再次查询实际上是基于 iframe 的 document
                所以这里在新建 iframe 的时候是要基于 iframe document 的 nodeId，这点和普通元素不一样
                
                不同域名的情况下，实际上就是另一个 page 去对待即可
            """
            if frame_id in json.dumps(frame_tree, ensure_ascii=False):  # 同域的话会获取到比如在 'childFrames' 内
                elements.append(Element(
                    node_id=frame_detail["node"]["contentDocument"]['nodeId'],
                    page=self,
                    is_iframe=True,
                    iframe_node_id=node_id,
                    iframe_owner_page=self
                ))
            else:
                try:
                    page_driver = PageDriver(
                        opts=self._opts,
                        page_id=frame_id,  # 页面的 targetId
                        context_id=self.context_id,  # 页面的上下文 id
                        context_driver=self._context_driver,
                        proxy=self._proxy,
                    )
                    status, iframe_dom = page_driver.cdp.dom.get_document()
                    if not status:
                        continue
                    elements.append(Element(
                        node_id=iframe_dom['root']['nodeId'],  # 这里的原理和上面一样，直接放 dom 的 id
                        page=page_driver,
                        is_iframe=True,
                        iframe_node_id=node_id,
                        iframe_owner_page=self
                    ))
                except websocket.WebSocketBadStatusException:
                    logger.warning("当前 iframe 可能已不在页面上：{}", frame_id)

        return elements

    def iframe_search(self, attr_name: str, attr_value: str, is_regex: bool = False) -> "Element":
        """
        这里是提供了一组可以通过指定属性，通过【正则、in】去匹配属性值的方法

        :param attr_name: 需要去判断的属性名
        :param attr_value: 对应属性名
        :param is_regex: 为 False 时通过 in 判断，为 True 时通过 re 正则判断
        :return:
        """
        iframes = self.iframes()

        for iframe in iframes:
            if not iframe.attrs.get(attr_name):
                continue

            attr = iframe.attrs[attr_name]

            if is_regex:
                if re.search(attr_value, attr):
                    return iframe
            elif attr_value in attr:
                return iframe

    def cookies(self, only_key_val: bool = True, domain: str = None) -> Union[List[dict], dict]:
        """
        获取 cookie

        :param only_key_val: 是否仅返回 {name:'',value:''} 形式
        :param domain: 指定域名
        :return: [{'name': '', 'value': '', 'domain': '', 'path': '', 'expires': -1, 'size': 161, 'httpOnly': True, 'secure': True, 'session': True, 'sameSite': 'Strict', 'priority': 'Medium', 'sameParty': False, 'sourceScheme': 'Secure', 'sourcePort': 443}]
        """
        status, cookies = self.cdp.storage.get_cookies(context_id=self.context_id)
        if not status:
            return []

        cookies = cookies['cookies']

        if only_key_val:
            cookie_dict = {}
            for c in cookies:
                if domain:
                    if c['domain'] == domain:
                        cookie_dict[c['name']] = c['value']
                else:
                    cookie_dict[c['name']] = c['value']

            return cookie_dict
        else:
            if domain:
                return list(filter(lambda x: x['domain'] == domain, cookies))
            return cookies

    def screenshot(
            self,
            save_path: str = None,
            format_type: str = "png",
            quality: int = None,
            clip: dict = None,
            from_surface: bool = True,
            capture_beyond_viewport: bool = False,
            optimize_for_speed: bool = False,
    ) -> bytes:
        """
        截图

        https://chromedevtools.github.io/devtools-protocol/tot/Page/#method-captureScreenshot

        clip 参考：https://chromedevtools.github.io/devtools-protocol/tot/Page/#type-Viewport

        :param save_path: 保存本地的路径
        :param format_type: jpeg, png, webp，默认值为 png
        :param quality: 压缩质量范围为 [0..100]（仅限 jpeg）
        :param clip: 仅捕获给定区域的屏幕截图（{x:1,y:1,width:1,height:1,scale:1}）
        :param from_surface: 从表面捕获屏幕截图，而不是从视图捕获屏幕截图，默认值为 true
        :param capture_beyond_viewport: 捕获视口之外的屏幕截图，默认值为 false
        :param optimize_for_speed: 优化图像编码以提高速度，而不是优化结果大小（默认为 false）
        :return: {"data":""}
        """
        # 等待页面加载
        self.wait_loaded()

        # 截图
        status, data = self.cdp.page.capture_screenshot(
            format_type=format_type,
            quality=quality,
            clip=clip,
            from_surface=from_surface,
            capture_beyond_viewport=capture_beyond_viewport,
            optimize_for_speed=optimize_for_speed
        )

        if not status or not data:
            return b""

        data_bytes = base64.b64decode(data["data"])

        # 保存本地
        if save_path:
            save_path = Path(save_path).absolute().resolve()
            save_path.parent.mkdir(parents=True, exist_ok=True)
            if not save_path.suffix:
                save_path = save_path.with_suffix(".png")

            with open(save_path, "wb") as f:
                f.write(data_bytes)

        return data_bytes

    def reload(
            self,
            ignore_cache: bool = False,
            scrip_to_evaluate_onload: str = None,
            raise_err: bool = False,
            need_callback: bool = False,
    ) -> Tuple[bool, dict]:
        """
        刷新页面

        https://chromedevtools.github.io/devtools-protocol/tot/Page/#method-reload

        :param ignore_cache: 是否忽略浏览器缓存(所有资源都会重新加载)
        :param scrip_to_evaluate_onload：注入的脚本（origin 变了就不会注入了）
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.cdp.page.reload(
            ignore_cache=ignore_cache,
            scrip_to_evaluate_onload=scrip_to_evaluate_onload,
            raise_err=raise_err,
            need_callback=need_callback
        )

    def stop_loading(
            self,
            raise_err: bool = False,
            need_callback: bool = False,
    ) -> Tuple[bool, dict]:
        """
        停止页面加载（强制页面停止所有导航和挂起的资源提取）

        https://chromedevtools.github.io/devtools-protocol/tot/Page/#method-stopLoading

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.cdp.page.stop_loading(
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def close(self):
        """
        关闭当前页面

        :return:
        """
        self.cdp.page.close()

    """
        等待
    """

    def wait_loaded(self, status: str = "complete", timeout: int = None):
        """
        等待页面加载完成

        :param status: 可选 interactive、complete
        :param timeout: 没有提供会使用全局的
        :return:
        """
        timeout = timeout or self._opts.timeout

        # 输入检查
        dom_read_states = ["interactive", "complete"]
        if status not in dom_read_states:
            status = "complete"

        # 加载等待
        start_time = time.time()
        while time.time() - start_time < timeout:
            dom_status = self.ready_state
            if dom_status == status or dom_status == "complete":
                self.__loading = False
                return
            time.sleep(0.1)

    def wait_for_selector(
            self,
            selector: str,
            node_id: int = None,
            timeout: int = None,
            raise_err: bool = False
    ) -> Union['Element', None, errors.ElementWaitTimeout]:
        """
        等待某个元素

        :param timeout:
        :param node_id:
        :param selector:
        :param raise_err:
        :return:
        """
        timeout = timeout or self._opts.timeout

        start_time = time.time()

        while time.time() - start_time < timeout:
            ele = self.query_selector(selector=selector, node_id=node_id)
            if ele:
                return ele
            time.sleep(0.1)

        if raise_err:
            raise errors.ElementWaitTimeout(selector)

    def wait_for_selectors(
            self,
            selectors: List[str],
            node_id: int = None,
            timeout: int = None,
            raise_err: bool = False
    ) -> Union['Element', None, errors.ElementWaitTimeout]:
        """
        等待多个元素，谁先拿到就返回谁（所以要自己注意放入的顺序）

        :param timeout:
        :param node_id:
        :param selectors: 元素列表
        :param raise_err:
        :return:
        """
        timeout = timeout or self._opts.timeout

        start_time = time.time()

        while time.time() - start_time < timeout:
            for s in selectors:
                ele = self.query_selector(selector=s, node_id=node_id)
                if ele:
                    return ele

                if time.time() - start_time >= timeout:
                    break

            time.sleep(0.1)

        if raise_err:
            raise errors.ElementWaitTimeout(selectors)

    """
        访问属性
    """

    @property
    def user_agent(self) -> str:
        """
        返回 ua

        :return:
        """
        status, data = self.cdp.browser.version()
        if not status:
            return ""
        return data["userAgent"]

    @property
    def ready_state(self) -> str:
        """
        页面加载状态

        :return:
        """
        return self.eval_js(js="document.readyState")[-1]

    @property
    def url(self) -> str:
        """
        当前页面的 url

        :return:
        """
        status, root = self.cdp.dom.get_document()

        if not status:
            return ""

        return root['root']['documentURL']

    @property
    def content(self) -> str:
        """
        当前页面的内容

        :return:
        """
        status, root = self.cdp.dom.get_document()
        if not status:
            return ""

        back_end_node_id = root['root']['backendNodeId']

        status, html = self.cdp.dom.get_outer_html(backendNodeId=back_end_node_id)
        if not status:
            return ""

        return html["outerHTML"]

    @property
    def listen(self) -> Listener:
        """
        事件监听

        :return:
        """
        return self._context_driver.listen

    """
       事件监听
    """

    @other_utils.catch_error
    def __on_page_life_cycle_event(self, *args, **kwargs):
        """
        针对顶级页面生命周期事件（如导航、加载、绘制等）触发

        :return:
        """
        self.__loading = True
        logger.debug("页面开始加载：{}", kwargs)

    @other_utils.catch_error
    def __on_page_load_event_fired(self, *args, **kwargs):
        """
        页面加载完成将会被触发

        :return:
        """
        self.__loading = False
        logger.debug("页面加载耗时：{}", kwargs['timestamp'])

    @other_utils.catch_error
    def __on_page_frame_started_loading(self, *args, **kwargs):
        """
        加载开始

        :return:
        """
        self.__loading = True

    @other_utils.catch_error
    def __on_page_frame_stopped_loading(self, *args, **kwargs):
        """
        加载结束

        :return:
        """
        self.__loading = False

    @other_utils.catch_error
    def __on_fetch_request_paused(self, *args, **kwargs):
        """
        请求监听

        请求将暂停，直到客户端响应 替换为 continueRequest、failRequest 或 fulfillRequest

        https://chromedevtools.github.io/devtools-protocol/tot/Fetch/#event-requestPaused

        :param args:
        :param kwargs: {"method":"Fetch.requestPaused","params":{"requestId":"interception-job-1.0","request":{"url":"https://www.baidu.com/","method":"GET","headers":{"Accept":"text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7","Upgrade-Insecure-Requests":"1","User-Agent":"Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/125.0.0.0 Safari/537.36","sec-ch-ua":"\"Chromium\";v=\"125\", \"Not.A/Brand\";v=\"24\"","sec-ch-ua-mobile":"?0","sec-ch-ua-platform":"\"Windows\""},"initialPriority":"VeryHigh","referrerPolicy":"strict-origin-when-cross-origin"},"frameId":"E361A685E14EAEC054BEF17592DF25DF","resourceType":"Document"}}
        :return:
        """
        request_id = kwargs["requestId"]
        network_id = kwargs.get("networkId")  # fetch 的 networkId 就是 network 的 requestId

        self.cdp.fetch.continue_request(
            request_id=request_id,
        )

        if not network_id:
            return

        # 请求输出
        request = Request(kwargs)
        # logger.warning("requests 存入：{}", network_id)
        self.__requests_cache[network_id] = request

        # 请求触发
        self.listen._on_request(request=request)

    @other_utils.catch_error
    def __on_network_request_will_be_sent(self, *args, **kwargs):
        """
        请求监听

        有时候 fetch 监听不到，需要通过这里监听

        https://chromedevtools.github.io/devtools-protocol/tot/Network/#event-requestWillBeSent

        :param args:
        :param kwargs:
        :return:
        """
        request_id = kwargs['requestId']

        if request_id not in self.__requests_cache:
            # logger.warning("request_will_be_sent 存入：{}", request_id)

            request = Request(kwargs)

            self.__requests_cache[request_id] = request

            # 请求触发
            self.listen._on_request(request=request)

    @other_utils.catch_error
    def __on_network_request_will_be_sent_extra_info(self, *args, **kwargs):
        """
        请求监听

        https://chromedevtools.github.io/devtools-protocol/tot/Network/#event-requestWillBeSentExtraInfo

        :param args:
        :param kwargs:
        :return:
        """
        request_id = kwargs['requestId']

    @other_utils.catch_error
    def __on_fetch_auth_required(self, *args, **kwargs):
        """
        监听 authRequired

        https://chromedevtools.github.io/devtools-protocol/tot/Fetch/#event-authRequired

        :param args:
        :param kwargs:
        :return:
        """
        request_id = kwargs['requestId']

        user = None
        pwd = None
        if self._proxy:
            user = self._proxy.get("user")
            pwd = self._proxy.get("pwd")
        elif self._opts.proxy and "user" in self._opts.proxy and "pwd" in self._opts.proxy:
            user = self._opts.proxy.get("user")
            pwd = self._opts.proxy.get("pwd")

        if not user or not pwd:
            logger.error("未获取到代理的账号密码！")
            return

        self.cdp.fetch.continue_with_auth(
            request_id=request_id,
            user=user,
            pwd=pwd,
        )

    @other_utils.catch_error
    def __on_network_response_received(self, *args, **kwargs):
        """
        响应监听

        https://chromedevtools.github.io/devtools-protocol/tot/Network/#event-responseReceived

        :param args:
        :param kwargs:
        :return:
        """
        request_id = kwargs["requestId"]

        # 存储响应
        self.__response_cache[request_id] = kwargs

    @other_utils.catch_error
    def __on_network_loading_finished(self, *args, **kwargs):
        """
        响应结束

        https://chromedevtools.github.io/devtools-protocol/tot/Network/#event-loadingFinished

        :param args:
        :param kwargs:
        :return:
        """
        request_id = kwargs["requestId"]

        # 获取请求
        if request_id not in self.__requests_cache:
            logger.warning("请求缓存无该 id：{}", request_id)
            return
        request_cache = self.__requests_cache[request_id]
        response_cache = self.__response_cache[request_id]

        # 存储输出
        response = Response(response_cache, request_cache, self.cdp)

        # 触发响应
        self.listen._on_response(response, self.cdp)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


"""
    动作链
"""


# modifiers
class Key:
    Alt = 1
    Ctrl = 2
    Backspace = 8
    Shift = 4
    Meta = 8


# windowsVirtualKeyCode
class KeyCode:
    Backspace = 8
    Tab = 9
    Enter = 13
    Shift = 16
    Ctrl = 17
    Alt = 18
    PauseOrBreak = 19
    CapsLock = 20
    Esc = 27
    Space = 32
    PageUp = 33
    PageDown = 34
    End = 35
    Home = 36
    LeftArrow = 37
    UpArrow = 38
    RightArrow = 39
    DownArrow = 40
    Insert = 45
    Delete = 46
    A = 65
    B = 66
    C = 67
    D = 68
    E = 69
    F = 70
    G = 71
    H = 72
    I = 73
    J = 74
    K = 75
    L = 76
    M = 77
    N = 78
    O = 79
    P = 80
    Q = 81
    R = 82
    S = 83
    T = 84
    U = 85
    V = 86
    W = 87
    X = 88
    Y = 89
    Z = 90
    Windows = 91  # Windows 键
    RightClick = 93
    Numpad_0 = 96
    Numpad_1 = 97
    Numpad_2 = 98
    Numpad_3 = 99
    Numpad_4 = 100
    Numpad_5 = 101
    Numpad_6 = 102
    Numpad_7 = 103
    Numpad_8 = 104
    Numpad_9 = 105
    Multiply = 106
    Add = 107
    Subtract = 109
    DecimalPoint = 110
    Divide = 111
    F1 = 112
    F2 = 113
    F3 = 114
    F4 = 115
    F5 = 116
    F6 = 117
    F7 = 118
    F8 = 119
    F9 = 120
    F10 = 121
    F11 = 122
    F12 = 123
    NumLock = 144
    ScrollLock = 145
    SemiColon = 186
    EqualSign = 187
    Comma = 188
    Dash = 189
    Period = 190
    ForwardSlash = 191
    GraveAccent = 192
    OpenBracket = 219
    BackSlash = 220
    CloseBracket = 221
    SingleQuote = 222


class ActionChain:
    def __init__(self, node_id: int, page: PageDriver, ele: "Element"):
        self.page = page
        self.node_id = node_id
        self.ele = ele

    def focus(self):
        """
        聚焦

        :return:
        """
        self.page.cdp.dom.focus(nodeId=self.node_id)

    def clear(self, simulation: bool = True):
        """
        清空输入框的值

        :param simulation: 模拟真实操作
        :return:
        """
        if not simulation:
            # 非模拟的话，实际的修改可能并不会生效（元素修改了，但是页面没更改）
            self.page.cdp.dom.set_attribute_value(
                name="value",
                value="",
                node_id=self.node_id,
            )
        else:
            # 全选
            self.ctrl(key=KeyCode.A)

            # 删除
            self.page.cdp.input.dispatch_key_event(
                event_type="rawKeyDown",
                windowsVirtualKeyCode=KeyCode.Backspace,
                modifiers=Key.Backspace,
            )

        return self

    def input_text(self, val: str, delay: float = 0, simulation: bool = True):
        """
        输入文本（模拟输入）

        :param val:
        :param simulation: 模拟真实操作
        :param delay: 延时输入（s）
        :return:
        """
        if not simulation:
            # 非模拟的话，实际的修改可能并不会生效（元素修改了，但是页面没更改）
            self.page.cdp.dom.set_attribute_value(
                name="value",
                value=val,
                node_id=self.node_id,
            )
        else:
            # 对焦
            self.focus()

            # 输入
            for v in val:
                self.page.cdp.input.dispatch_key_event(
                    event_type="keyDown",
                    is_key_pad=True,
                    key=v,
                    text=v,
                    code="Key " + v.upper(),
                )

                self.page.cdp.input.dispatch_key_event(
                    event_type="keyUp",
                    is_key_pad=True,
                    key=v,
                    text=v,
                    code="Key " + v.upper(),
                )

                time.sleep(delay)

        return self

    def ctrl(self, key: int):
        """
        ctrl 事件

        :param key: 如 ctrl+A 只要输入 a 或 A 就行了
        :return:
        """
        # 对焦
        self.focus()

        # 输入
        self.page.cdp.input.dispatch_key_event(
            event_type="rawKeyDown",
            windowsVirtualKeyCode=key,
            modifiers=Key.Ctrl,
        )

        return self

    # TODO 增加组合输入
    # def combination_input(self, modifiers: int, value: Union[str, int] = None):
    #     """
    #     组合输入
    #
    #     :param modifiers: 比如 ctrl+shift 就是 Key.Ctrl|Key.Shift
    #     :param value: 输入的值 比如 abc
    #     :return:
    #     """

    def click(self, count: int = 1, button: str = "left", simulation: bool = True):
        """
        执行点击

        :param count: 点击次数
        :param button: 代表鼠标的按键，如：none（不触发按键，代表光标）, left（左击）, middle（滚轮）, right（右击）, back（后退）, forward（前进）
        :param simulation: 模拟真实操作
        :return:
        """
        if not simulation:
            self.ele.eval_js_on_selector(js="this.click();", returnByValue=True)
        else:
            # 聚焦
            self.focus()

            # 获取元素坐标
            x, y = self.ele.coordinate_center

            # 移动
            self.page.cdp.input.dispatch_mouse_event(
                event_type="mouseMoved",
                x=x,
                y=y,
                button="none"
            )

            # 点击
            self.page.cdp.input.dispatch_mouse_event(
                event_type="mousePressed",
                x=x,
                y=y,
                button=button,
                clickCount=count
            )

            # 释放
            self.page.cdp.input.dispatch_mouse_event(
                event_type="mouseReleased",
                x=x,
                y=y,
                button=button,
                clickCount=count
            )

        return self


"""
    元素
"""


class Element(ActionChain):
    def __init__(
            self,
            node_id: int,
            page: PageDriver,
            is_iframe: bool = False,
            iframe_node_id: int = None,
            iframe_owner_page: PageDriver = None
    ):
        """
        生成节点

        如果是 iframe 那么这里获取的实际上就是 iframe 的 dom 而不是 iframe 自身
        :param node_id 当前节点的 id（iframe 的话就是 iframe document 的 node_id）
        :param page PageDriver 对象（iframe 的话就是 iframe 自身的 driver）
        :param is_iframe 是不是 iframe
        :param iframe_node_id iframe 自身的 node_id
        :param iframe_owner_page iframe 所在页面
        :return:
        """
        if is_iframe:
            super().__init__(iframe_node_id, iframe_owner_page, self)
        else:
            super().__init__(node_id, page, self)

        self._back_end_node_id = None
        self._obj_id = None
        self._is_iframe = is_iframe
        self._iframe_node_id = iframe_node_id
        self._iframe_owner_page = iframe_owner_page
        self._iframe_page = page
        self._iframe_document_node_id = node_id

        self.__tagName: str = None if not is_iframe else "iframe"  # type:ignore

    def query_selector(
            self,
            selector: str,
            raise_err: bool = False
    ) -> Union['Element', None, errors.ElementNotFound]:
        """
        选择元素

        通过这种绕一圈的书写方式，就可以有编辑器的提示了

        :param selector: 选择器（支持 css、xpath）
        :param raise_err: 是否报错
        :return:
        """
        if self._is_iframe:
            return self._iframe_page.query_selector(
                selector=selector,
                node_id=self._iframe_document_node_id,
                raise_err=raise_err
            )
        return self.page.query_selector(
            selector=selector,
            node_id=self.node_id,
            raise_err=raise_err
        )

    def query_selector_all(
            self,
            selector: str,
            raise_err: bool = False
    ) -> Union['Element', None, errors.ElementNotFound]:
        """
        选择元素

        :param selector: 选择器（支持 css、xpath）
        :param raise_err: 是否报错
        :return:
        """
        if self._is_iframe:
            return self._iframe_page.query_selector_all(
                selector=selector,
                node_id=self._iframe_document_node_id,
                raise_err=raise_err
            )
        return self.page.query_selector_all(
            selector=selector,
            node_id=self.node_id,
            raise_err=raise_err,
        )

    def wait_for_selector(
            self,
            selector: str,
            timeout: int = None,
            raise_err: bool = False
    ) -> Union['Element', None, errors.ElementWaitTimeout]:
        """
        等待某个元素

        :param timeout:
        :param selector:
        :param raise_err:
        :return:
        """
        if self._is_iframe:
            return self._iframe_page.wait_for_selector(
                selector=selector,
                timeout=timeout,
                node_id=self._iframe_document_node_id,
                raise_err=raise_err
            )
        return self.page.wait_for_selector(
            selector=selector,
            timeout=timeout,
            node_id=self.node_id,
            raise_err=raise_err
        )

    def wait_for_selectors(
            self,
            selectors: List[str],
            timeout: int = None,
            raise_err: bool = False
    ) -> Union['Element', None, errors.ElementWaitTimeout]:
        """
        等待多个元素，谁先拿到就返回谁（所以要自己注意放入的顺序）

        :param timeout:
        :param selectors: 元素列表
        :param raise_err:
        :return:
        """
        if self._is_iframe:
            return self._iframe_page.wait_for_selectors(
                selectors=selectors,
                timeout=timeout,
                node_id=self._iframe_document_node_id,
                raise_err=raise_err
            )
        return self.page.wait_for_selectors(
            selectors=selectors,
            timeout=timeout,
            node_id=self.node_id,
            raise_err=raise_err
        )

    def eval_js_on_selector(self, js: str, **kwargs) -> Tuple[bool, dict]:
        """
        执行 js 脚本（会自动使用 function 包裹，不然无法执行）

        学习链接：https://chromedevtools.github.io/devtools-protocol/tot/Runtime/#method-evaluate

        :param js: 路径或者 js 字符串
        :return:
        """
        js = "function (){ %s }" % js
        return self.page.cdp.runtime.call_function_on(js=js, object_id=self.object_id, **kwargs)

    @property
    def style(self) -> dict:
        """
        获取 style

        :return:
        """
        status, content = self.page.cdp.css.get_computed_style_for_node(node_id=self.node_id)
        if not status:
            return {}

        return dict([[i["name"], i["value"]] for i in content["computedStyle"]])

    @property
    def is_visible(self) -> bool:
        """
        元素是否可见（全部可见）

        :return:
        """
        js = """
            return function (el) {
                var rect = el.getBoundingClientRect();
                var isVisible = (
                    rect.top >= 0 &&
                    rect.left >= 0 &&
                    rect.bottom <= (window.innerHeight || document.documentElement.clientHeight) &&
                    rect.right <= (window.innerWidth || document.documentElement.clientWidth)
                );
            
                if (!isVisible) return false;
            
                var style = window.getComputedStyle(el);
                if (style.visibility === 'hidden' || style.display === 'none') return false;
            
                if (rect.width === 0 || rect.height === 0) return false;
            
                var elementBelow = document.elementFromPoint(rect.left + rect.width / 2, rect.top + rect.height / 2);
                if (elementBelow && elementBelow !== el && !el.contains(elementBelow)) return false;
            
                return true;
            }(this)
        """
        status, val = self.eval_js_on_selector(js=js)
        if not status:
            return False

        return val["result"]["value"]

    @property
    def is_displayed(self) -> bool:
        """
        元素是否显示

        :return:
        """
        style = self.style

        if style.get("visibility") == 'hidden':
            return True

        if style.get("display") == 'none':
            return True

        if self.attrs.get("hidden"):
            return True

        status, val = self.eval_js_on_selector(js="return this.offsetParent === null;")
        if status and val["result"]["value"]:
            return True

        return False

    @property
    def is_enable(self) -> bool:
        """
        元素是否可用

        :return:
        """
        status, val = self.eval_js_on_selector(js="return this.disabled;")
        if not status:
            return False
        return val["result"]["value"]

    @property
    def is_checked(self) -> bool:
        """
        元素是否被选择

        :return:
        """
        status, val = self.eval_js_on_selector(js="return this.checked;")
        if not status:
            return False
        return val["result"]["value"]

    @property
    def is_selected(self) -> bool:
        """
        元素是否被选择

        :return:
        """
        status, val = self.eval_js_on_selector(js="return this.selected;")
        if not status:
            return False
        return val["result"]["value"]

    @property
    def coordinates(self) -> dict:
        """
        获取元素的坐标信息
        返回的比如 content 有 4 个坐标，分别是
        - 左上顶点
        - 右上顶点
        - 右下顶点
        - 左下顶点

        :return:
        """
        # 获取元素坐标
        status, content = self.page.cdp.dom.get_box_model(nodeId=self.node_id)
        if not status:
            return {}

        return content['model']

    @property
    def coordinate_center(self) -> tuple:
        """
        元素中心点坐标

        :return:
        """
        content = self.coordinates["content"]
        return browser_utils.parse_box_center_coordinate(data=content)

    @property
    def attrs(self) -> dict:
        """
        属性

        :return:
        """
        status, attrs = self.page.cdp.dom.get_attributes(node_id=self.node_id)

        if not status:
            return {}

        attrs_list = attrs["attributes"]
        attrs = dict(zip(attrs_list[::2], attrs_list[1::2]))

        return attrs or {}

    @property
    def tag_name(self) -> str:
        """
        节点名称

        :return:
        """
        if not self.__tagName:
            status, describe = self.page.cdp.dom.describe_node(nodeId=self.node_id)
            if status and describe:
                self.__tagName = describe['node']['nodeName'].lower()
        return self.__tagName

    @property
    def html(self) -> str:
        """
        html 文本

        iframe 获取的是 iframe document 的 html

        :return: <input type="submit" id="su" value="百度一下" class="bg s_btn">
        """
        if self._is_iframe:
            describe_status, node = self._iframe_page.cdp.dom.describe_node(nodeId=self._iframe_document_node_id)
            if not describe_status:
                return ""
            iframe_document_back_end_node_id = node['node']['backendNodeId']
            status, html = self._iframe_page.cdp.dom.get_outer_html(backendNodeId=iframe_document_back_end_node_id)
        else:
            status, html = self.page.cdp.dom.get_outer_html(backendNodeId=self.back_end_node_id)

        if not status:
            return ""

        return html["outerHTML"]

    @property
    def text(self):
        """
        获取除了标签以外的内容

        :return:
        """
        return self.soup.get_text()

    @property
    def selector(self) -> Selector:
        """
        获取 parsel 解析器

        :return:
        """
        return Selector(self.html)

    @property
    def soup(self) -> BeautifulSoup:
        """
        获取 bs4 解析器

        :return:
        """
        return BeautifulSoup(self.html, "lxml")

    @property
    def back_end_node_id(self) -> int:
        """
        获取 backendNodeId

        :return:
        """
        if not self._back_end_node_id:
            _, node = self.page.cdp.dom.describe_node(nodeId=self.node_id)
            self._back_end_node_id = node['node']['backendNodeId']

        return self._back_end_node_id

    @property
    def object_id(self) -> str:
        """
        获取 object_id

        :return:
        """
        if not self._obj_id:
            _, node = self.page.cdp.dom.resolve_node(backendNodeId=self.back_end_node_id)
            self._obj_id = node['object']['objectId']

        return self._obj_id

    def __repr__(self):
        attrs = self.attrs
        attrs_text = ' '.join([f'{i}="{attrs[i]}"' for i in attrs.keys()])
        return f"<{self.__class__.__name__} {self.tag_name} {attrs_text}>"



import base64
from typing import Union

from easypage.cdp import CDP
from easypage.request import Request


class Response:

    def __init__(self, data: dict, request: Request, cdp: "CDP"):
        """

        :param data: 响应体
        :param request: 请求
        :param cdp: cdp
        """
        self.__cdp = cdp
        self.request = request
        self.request_id = data["requestId"]
        self._response = data["response"]
        self.response_from = data["type"]

        # 发出获取响应体的请求（这个请求不能立刻获取结果不然会卡住，而且可能也没有加载完成）
        self._body: Union[str, bytes] = ""  # 暂时为空

    def _get_body(self):
        """
        获取响应体

        :return:
        """
        get_status, body_await = self.__cdp.network.get_response_body(
            request_id=self.request_id,
            need_callback=False
        )

        def body_callback(status: bool, body_data: dict):
            """
            body 获取到之后重新设置

            :param status: 回调是否正常
            :param body_data:
            :return:
            """
            # 根据结果重新设置 body
            if status:
                body = body_data.get("body", b"")

                if body_data.get('base64Encoded'):
                    body = base64.b64decode(body)

                self._body = body

            # 触发事件
            self.__cdp.emit(event=f"handle_response_{self.request_id}")

        # 发送成功的话，创建事件
        if get_status:
            self.__cdp.once(
                event=f"Network.getResponseBody-{body_await['id']}",
                callback=body_callback
            )

    @property
    def url(self) -> str:
        return self._response["url"]

    @property
    def status(self) -> int:
        return self._response.get("status", 0)

    @property
    def status_text(self) -> str:
        """
        状态字符串：'OK'

        :return:
        """
        return self._response.get("statusText", 0)

    @property
    def headers(self) -> dict:
        return self._response.get("headers", {})

    @property
    def mime_type(self) -> str:
        """
        类型如：'text/html'

        :return:
        """
        return self._response.get("mimeType", "")

    @property
    def charset(self) -> str:
        """
        编码格式如：'charset'

        :return:
        """
        return self._response.get("charset", "")

    @property
    def remote_ip_address(self) -> str:
        """
        远程端口

        :return:
        """
        return self._response.get("remoteIPAddress", "")

    @property
    def protocol(self) -> str:
        """
        协议：'http/1.1'

        :return:
        """
        return self._response.get("protocol", "")

    @property
    def security_details(self) -> dict:
        """
        安全详情

        返回值大概如下：
            {
                "protocol": "TLS 1.3",
                "keyExchange": "",
                "keyExchangeGroup": "X25519",
                "cipher": "AES_256_GCM",
                "certificateId": 0,
                "subjectName": "*.csdnimg.cn",
                "sanList": [
                    "*.csdnimg.cn",
                    "csdnimg.cn"
                ],
                "issuer": "RapidSSL Global TLS RSA4096 SHA256 2022 CA1",
                "validFrom": 1696809600,
                "validTo": 1728777599,
                "signedCertificateTimestampList": [],
                "certificateTransparencyCompliance": "unknown",
                "serverSignatureAlgorithm": 2052,
                "encryptedClientHello": false
            }

        :return:
        """
        return self._response.get("securityDetails", {})

    @property
    def body(self) -> Union[str, bytes]:
        """
        响应体

        大致的关系如下：
            响应类型                返回值类型
            image/png              bytes
            application/json       str

        :return:
        """
        return self._body

    def __repr__(self):
        return f"<Response status={self.status} url=\"{self.url}\">"



"""
    使用了 EventEmitter 作为事件管理

    # 创建事件监听案例
    async def intercept(kwargs) -> None:
        print(kwargs)
        interception_id = kwargs["interceptionId"]

        await page_conn.send(
            method="Network.continueInterceptedRequest",
            params={"interceptionId": interception_id},
        )

    page_conn.on("Network.requestIntercepted", lambda event: loop.create_task(intercept(event)))

    异步这里代码结构有问题，仅针对 send 方法
        事件监听，监听到事件之后将会触发 loop.run_until_complete 操作
        同时同步程序也会有 loop.run_until_complete 就会造成报错：This event loop is already running
        需要对代码结构做处理，如有统一的事件发送地址，或者使用类似队列的结构
"""
import json
import time
import typing
import asyncio
import threading
import traceback
import websocket
import websockets
from threading import Lock
from easypage import errors, settings
from pyee import EventEmitter
from easypage.logger import logger
from queue import SimpleQueue, Empty
from websockets.legacy import client

TYPE_PAGE = "page"
TYPE_BROWSER = "browser"

DRIVER_TYPE = typing.Union[
    TYPE_BROWSER,
    TYPE_PAGE
]


class ConnAsync(EventEmitter):
    def __init__(self, address: str, driver_type: DRIVER_TYPE, driver_id: str, loop: asyncio.AbstractEventLoop = None):
        """
        # 页面连接
        ws://127.0.0.1:9223/devtools/page/87B827CF2CAF710C997699F858467545

        # 浏览器连接
        ws://127.0.0.1:9223/devtools/browser/0909caac-09eb-4f48-a481-5de5decb3583

        :param loop: 事件循环（没传的话，会自动创建）
        :param address: 连接地址，如：127.0.0.1:9223
        :param driver_type: 控制类型，如：TYPE_PAGE
        :param driver_id: 对应 id，如：87B827CF2CAF710C997699F858467545
        """
        super().__init__()
        self.__loop = loop or asyncio.get_event_loop()
        self.address = address
        self.driver_id = driver_id
        self.driver_type = driver_type
        self.driver_url = f"ws://{address}/devtools/{driver_type}/{driver_id}"

        self.__lock = Lock()
        self.__id = 0
        self.__events = {}  # 存储事件
        self.__callbacks = {}  # 管理所有回调
        self.__connected = False  # 是否已经连接上（client.WebSocketCommonProtocol 是否已创建）
        self.__ws = self.new_conn()  # ws 连接
        self.__conn: client.WebSocketCommonProtocol = None  # type: ignore # 实际操作的连接
        self.__recv_fut = self.__loop.create_task(self.__recv())  # 异步消息接收

    def new_conn(self) -> client.Connect:
        """
        创建新的 ws 连接，使用的时候需要创建 client.WebSocketCommonProtocol

        :return:
        """
        return client.connect(
            uri=self.driver_url,
            max_size=None,
            loop=self.__loop,
            ping_interval=None,
            ping_timeout=None
        )

    def send(self, method: str, params: dict = None, raise_err: bool = False) -> typing.Tuple[bool, dict]:
        """
        同步发送消息

        使用：status, value = send(method="", raise_err=False, params={})

        返回的结果是有两个值：
            - status 表示当前消息是否有错误发生（raise_err 为 True 时直接抛出）
            - value 是对应浏览器返回的消息，是一个 dict

        :param method:
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param params:
        :return
        """
        # 获得结果
        callback = self.send_async(
            method=method,
            raise_err=raise_err,
            params=params
        )

        # 在同步上下文中运行异步函数
        status, value = self.__loop.run_until_complete(asyncio.wait_for(callback, timeout=10))

        return status, value

    def send_async(self, method: str, raise_err: bool = False, params: dict = None) -> typing.Awaitable:
        """
        异步发送消息，返回一个需要 await 等待的对象

        使用：status, value = send(method="", raise_err=False, params={})

        返回的结果是有两个值：
            - status 表示当前消息是否有错误发生（raise_err 为 True 时直接抛出）
            - value 是对应浏览器返回的消息，是一个 dict

        :param method:
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param params:
        :return
        """
        # 发送消息
        msg = {
            "id": self.__next_id,
            "method": method,
            "params": params or dict()
        }
        self.__loop.create_task(self.__send(msg))

        # 构建 Future 对象返回
        callback = self.__loop.create_future()
        self.__callbacks[msg["id"]] = callback
        callback.raise_err: bool = raise_err  # type: ignore
        callback.error: Exception = errors.CDPRecvMsgError()  # type: ignore
        callback.method: str = method  # type: ignore
        return callback

    async def __send(self, msg: dict) -> None:
        """
        消息发送

        :param msg:
        :return:
        """
        # 等待连接建立
        while not self.__connected:
            await asyncio.sleep(0)

        # 发送消息
        msg_id = msg["id"]
        msg = json.dumps(msg, ensure_ascii=False)
        logger.debug("[SEND] {}", msg)

        try:
            await self.__conn.send(msg)
        except (websockets.ConnectionClosed, websockets.ConnectionClosedOK):
            logger.warning(f"连接已关闭：{self.driver_url}")

            # 获取回调对象并设置结果
            callback = self.__callbacks.get(msg_id, None)
            if callback and not callback.done():
                callback.set_result((False, dict()))
                await self.close()

    async def __recv(self) -> None:
        """
        接收消息

        :return:
        """
        async with self.__ws as conn:
            self.__conn = conn
            self.__connected = True
            while self.__connected:
                try:
                    msg = await self.__conn.recv()
                    if not msg:
                        break
                    await self.__onmessage(msg=msg)
                except (websockets.ConnectionClosed, ConnectionResetError):
                    logger.warning(f"连接已关闭：{self.driver_url}")
                    break
                await asyncio.sleep(0)

    async def __onmessage(self, msg: str) -> None:
        """
        接收到的消息处理

        :param msg:
        :return:
        """
        await asyncio.sleep(0)
        if "Not allowed" in msg:
            logger.warning(f"当前 driver 不支持该命令，请切换 driver，当前 driver 类型：{self.driver_type}")
        msg = json.loads(msg)

        # 消息回调
        if msg.get('id') in self.__callbacks:
            callback = self.__callbacks.pop(msg['id'])
            if msg.get('error'):
                logger.warning("[RECV] {}", msg)
                if callback.raise_err:
                    callback.error.args = (f"接收到错误消息：{msg['error']}",)  # 设置错误消息
                    callback.set_exception(callback.error)
                else:
                    callback.set_result((False, msg.get('error')))
            else:
                logger.debug("[RECV] {}", msg)
                callback.set_result((True, msg.get('result')))

        # 事件触发
        else:
            logger.debug("[RECV] {}", msg)
            method = msg.get('method', '')
            params = msg.get('params', {})
            self.emit(method, **params)
            print("接收到事件 -> ", method, self.__events.get(method))
            # if self.__events.get(method):
            #     for f in self.__events[method]:
            #         f(**params)

    # def on(self, event: str, f: typing.Callable):
    #     print("event -> ", event)
    #     if event not in self.__events:
    #         self.__events[event] = set()
    #
    #     self.__events[event].add(f)

    async def close(self) -> None:
        """
        关闭连接

        :return:
        """
        self.__connected = False
        await asyncio.sleep(0)

        # 未获取的结果设置异常
        for cb in self.__callbacks.values():
            if cb.raise_err:
                cb.error.args = (f'连接关闭：{cb.method}',)  # 设置错误消息
                cb.set_exception(cb.error)
            else:
                cb.set_result((False, f'连接关闭：{cb.method}'))

        self.__callbacks.clear()

        # 关闭连接
        if self.__conn:
            await self.__conn.close()
            self.__conn = None

        # 取消等待
        if not self.__recv_fut.done():
            self.__recv_fut.cancel()

    @property
    def __next_id(self) -> int:
        """
        获取消息 id

        :return:
        """
        self.__lock.acquire(timeout=2)
        self.__id += 1
        self.__lock.release()
        return self.__id


class Conn(EventEmitter):

    def __init__(self, address: str, driver_type: DRIVER_TYPE, driver_id: str, timeout: int = None):
        """
        # 页面连接
        ws://127.0.0.1:9223/devtools/page/87B827CF2CAF710C997699F858467545

        # 浏览器连接
        ws://127.0.0.1:9223/devtools/browser/0909caac-09eb-4f48-a481-5de5decb3583

        :param address: 连接地址，如：127.0.0.1:9223
        :param driver_type: 控制类型，如：TYPE_PAGE
        :param driver_id: 对应 id，如：87B827CF2CAF710C997699F858467545
        :param timeout: 配置的超时时间
        """
        super().__init__()
        self.address = address
        self.driver_id = driver_id
        self.driver_type = driver_type
        self.driver_url = f"ws://{address}/devtools/{driver_type}/{driver_id}"
        self.timeout = timeout or settings.TIMEOUT

        self.__lock = threading.Lock()
        self.__id: int = 0
        self.__stop: bool = False
        self.__ws: WebSocket = None  # type:ignore
        self.__result_monitor = {}  # 存储结果
        self.__new_conn()

    def __new_conn(self):
        """
        与浏览器建立连接

        :return:
        """
        self.__ws = websocket.create_connection(
            url=self.driver_url,
            enable_multithread=False,
            suppress_origin=True
        )
        threading.Thread(target=self.__recv_result, daemon=True).start()

    def send(
            self,
            method: str,
            params: dict = None,
            raise_err: bool = False,
            need_callback: bool = True,
    ) -> typing.Tuple[bool, dict]:
        """
        同步发送消息

        使用：status, value = send(method="", raise_err=False, params={})

        返回的结果是有两个值：
            - status 表示当前消息是否有错误发生（raise_err 为 True 时直接抛出）
            - value 是对应浏览器返回的消息，是一个 dict

        :param method:
        :param params:
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return
        """
        # 发送消息
        status, msg = self.__send(
            method=method,
            params=params,
            need_callback=need_callback
        )

        if not status:
            return False, dict()
        elif not need_callback:
            return True, msg

        # 等待信息接收
        start_time = time.time()
        while not self.__stop:
            if self.closed():
                break

            try:
                recv_msg = self.__result_monitor[msg["id"]].get(timeout=0.1)  # 短等待，避免有问题卡着
                if "result" in recv_msg:
                    return True, recv_msg["result"]
                elif "error" in recv_msg:
                    if raise_err:
                        raise errors.CDPRecvErrorMsg(recv_msg["error"])
                    return False, recv_msg["error"]

                logger.error("没有结果：{}", recv_msg)
            except Empty:
                if time.time() - start_time >= self.timeout:
                    self.__remove_result(msg_id=msg["id"])
                    if raise_err:
                        raise errors.CDPRecvTimeout(msg)
                    logger.error("接收超时：{}", msg)
                    break

    def __send(
            self,
            method: str,
            params: dict = None,
            need_callback: bool = True,
    ) -> typing.Tuple[bool, dict]:
        """
        同步发送消息

        使用：status, value = send(method="", raise_err=False, params={})

        返回的结果是有两个值：
            - status 表示当前消息是否有错误发生（raise_err 为 True 时直接抛出）
            - value 是对应浏览器返回的消息，是一个 dict

        :param method:
        :param params:
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return
        """
        # 构建消息
        msg = {
            "id": self.__get_id,
            "method": method,  # 在此示例中启用网络事件
            "params": params
        }

        if need_callback:
            self.__result_monitor[msg["id"]] = SimpleQueue()

        msg_json = json.dumps(msg)
        logger.debug("[SEND] {}", msg_json)

        # 发送信息
        try:
            self.__ws.send(msg_json)
            return True, msg
        except (errors.CDPRecvTimeout, errors.CDPRecvErrorMsg):
            raise
        except (websocket.WebSocketConnectionClosedException, ConnectionResetError):
            # logger.warning("连接已断开：{}", self.driver_url)
            raise errors.CDPConnClosed(f"连接已断开：{self.driver_url}")
        except:
            traceback.print_exc()
            raise

        # return False, dict()

    def __recv_result(self):
        """
        接收消息、事件处理

        :return:
        """
        while not self.__stop:
            try:
                msg = self.__ws.recv()
                logger.debug("[RECV] {}", msg)
                msg = json.loads(msg)
            except websocket.WebSocketTimeoutException:
                continue
            except (websocket.WebSocketConnectionClosedException, ConnectionResetError):
                logger.warning("连接已断开：{}", self.driver_url)
                return
            except Exception:
                if self.__stop:
                    return
                else:
                    self.__stop = True
                    raise

            # 消息处理
            if msg.get("id"):
                has_error = False
                if msg.get("error"):
                    logger.error("[RECV] {}", msg["error"])
                    has_error = True

                # 消息 id 存在的话就是需要立刻回调
                if msg["id"] in self.__result_monitor:
                    self.__add_result(msg_id=msg["id"], result=msg)
                else:
                    # 判断响应-body 获取事件的触发
                    event_name = f"Network.getResponseBody-{msg['id']}"
                    if event_name in self._events:
                        if has_error:
                            self.emit(event=event_name, status=False, body_data=msg["error"])
                        else:
                            self.emit(event=event_name, status=True, body_data=msg["result"])

            elif "method" in msg:
                method = msg.get("method", "")
                params = msg.get("params", {})
                if method in self._events:
                    self.emit(event=method, **params)

    def closed(self) -> bool:
        """
        判断连接是否断开，同时做一点处理

        :return:
        """
        status = self.__ws.connected
        if not status:
            self.__stop = True

        return not status

    def close(self):
        """
        停止运行

        :return:
        """
        self.__stop = True
        self.__ws.close()

    def __add_result(self, msg_id: int, result: dict):
        """
        添加监控结果

        :param msg_id:
        :param result:
        :return:
        """
        if msg_id in self.__result_monitor:
            self.__result_monitor[msg_id].put(result)

    def __remove_result(self, msg_id: int):
        """
        移除监控

        :param msg_id:
        :return:
        """
        if msg_id in self.__result_monitor:
            del self.__result_monitor[msg_id]

    @property
    def __get_id(self):
        """
        获取一个消息 id

        :return:
        """
        self.__lock.acquire(timeout=2)
        self.__id += 1
        self.__lock.release()

        return self.__id

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



from typing import Tuple
from easypage.conn import Conn


class NetWork:
    """
        NetWork

        https://chromedevtools.github.io/devtools-protocol/tot/Network
    """

    def __init__(self, conn: Conn):
        self.__conn = conn

    def enable(
            self,
            raise_err: bool = False,
            need_callback: bool = False,
            **kwargs
    ) -> Tuple[bool, dict]:
        """
        启用网络跟踪，网络事件现在将传递到客户端。

        https://chromedevtools.github.io/devtools-protocol/tot/Network/#method-enable

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="Network.enable",
            params=kwargs,
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def set_extra_http_headers(
            self,
            headers: dict,
            raise_err: bool = False,
            need_callback: bool = False,
    ) -> Tuple[bool, dict]:
        """
        指定是否始终随此页面的请求发送额外的 HTTP 标头。

        https://chromedevtools.github.io/devtools-protocol/tot/Network/#method-setExtraHTTPHeaders

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :param headers:
        :return:
        """
        return self.__conn.send(
            method="Network.setExtraHTTPHeaders",
            params={
                "headers": headers
            },
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def set_user_agent_override(
            self,
            user_agent: str,
            raise_err: bool = False,
            need_callback: bool = False,
            **kwargs
    ) -> Tuple[bool, dict]:
        """
        允许使用给定的字符串覆盖用户代理。

        https://chromedevtools.github.io/devtools-protocol/tot/Network/#method-setUserAgentOverride

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :param user_agent:
        :return:
        """
        send_data = {
            'userAgent': user_agent
        }

        send_data.update(kwargs)

        return self.__conn.send(
            method="Network.setUserAgentOverride",
            params=send_data,
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def set_cache_disabled(
            self,
            cache_disabled: bool = True,
            raise_err: bool = False,
            need_callback: bool = False,
    ) -> Tuple[bool, dict]:
        """
        禁用缓存

        https://chromedevtools.github.io/devtools-protocol/tot/Network/#method-setCacheDisabled

        :param cache_disabled:
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="Network.setCacheDisabled",
            params={
                "cacheDisabled": cache_disabled
            },
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def get_response_body(
            self,
            request_id: str,
            raise_err: bool = False,
            need_callback: bool = True,
    ) -> Tuple[bool, dict]:
        """
        获取响应体

        这个方法会导致卡住，所以不能直接拿到结果

        https://chromedevtools.github.io/devtools-protocol/tot/Network/#method-getResponseBody

        :param request_id:
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="Network.getResponseBody",
            params={
                "requestId": request_id
            },
            raise_err=raise_err,
            need_callback=need_callback
        )



"""
    Driver 的基类
"""
from easypage.cdp import CDP
from abc import abstractmethod
from easypage.conn import Conn, DRIVER_TYPE
from easypage.options import BrowserOptions


class Driver:
    def __init__(self, opts: BrowserOptions, driver_type: DRIVER_TYPE, driver_id: str):
        """

        :param opts:
        :param driver_type:
        :param driver_id:
        """
        self._opts = opts
        self.__conn = Conn(
            address=opts.address,
            driver_type=driver_type,
            driver_id=driver_id,
            timeout=opts.timeout
        )
        self.__cdp = CDP(conn=self.__conn)

    @property
    def cdp(self) -> CDP:
        """
        获取 Chrome DevTools Protocol 实例

        :return:
        """
        return self.__cdp

    @abstractmethod
    def close(self):
        """
        资源回收

        :return:
        """



from typing import Tuple
from urllib import parse
from easypage import errors
from easypage.conn import Conn


class Page:
    """
        Page

        https://chromedevtools.github.io/devtools-protocol/tot/Page
    """

    def __init__(self, conn: Conn):
        self.__conn = conn

    def enable(
            self,
            raise_err: bool = False,
            need_callback: bool = False,
    ) -> Tuple[bool, dict]:
        """
        启用页面功能（加载和管理页面、导航、截图等）

        https://chromedevtools.github.io/devtools-protocol/tot/Page/#method-enable

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="Page.enable",
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def reload(
            self,
            ignore_cache: bool = False,
            scrip_to_evaluate_onload: str = None,
            raise_err: bool = False,
            need_callback: bool = False,
    ) -> Tuple[bool, dict]:
        """
        刷新页面

        https://chromedevtools.github.io/devtools-protocol/tot/Page/#method-reload

        :param ignore_cache: 是否忽略浏览器缓存(所有资源都会重新加载)
        :param scrip_to_evaluate_onload：注入的脚本（origin 变了就不会注入了）
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="Page.enable",
            params={
                "ignoreCache": ignore_cache,
                "scriptToEvaluateOnLoad": scrip_to_evaluate_onload,
            },
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def stop_loading(
            self,
            raise_err: bool = False,
            need_callback: bool = False,
    ) -> Tuple[bool, dict]:
        """
        停止页面加载（强制页面停止所有导航和挂起的资源提取）

        https://chromedevtools.github.io/devtools-protocol/tot/Page/#method-stopLoading

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="Page.stopLoading",
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def add_script_to_evaluate_on_new_document(
            self,
            source: str,
            run_immediately: bool = False,
            raise_err: bool = False,
            need_callback: bool = False,
    ) -> Tuple[bool, dict]:
        """
        每次加载新文档时执行脚本（在页面所有 scripts 执行之前执行，可以用作 hook 之类）

        https://chromedevtools.github.io/devtools-protocol/tot/Page/#method-addScriptToEvaluateOnNewDocument

        :param source: 执行的脚本
        :param run_immediately: 是否立即执行（否的话只在新文档加载时执行）
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        send_data = {
            "source": source,
        }
        if run_immediately:
            send_data["runImmediately"] = run_immediately

        return self.__conn.send(
            method="Page.addScriptToEvaluateOnNewDocument",
            params=send_data,
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def navigate(
            self,
            url: str,
            frame_id: str = None,
            ignore_url_check: bool = False,
            raise_err: bool = False,
            need_callback: bool = True,
            **kwargs
    ) -> Tuple[bool, dict]:
        """
        访问链接或本地文件

        https://chromedevtools.github.io/devtools-protocol/tot/Page/#method-navigate

        :param url:
        :param frame_id:
        :param ignore_url_check: 忽略 url 检查
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return: 错误的时候会返回错误的原因
        """
        # url 检查
        if not ignore_url_check:
            url_parse = parse.urlparse(url)
            if not url_parse.scheme or not url_parse.hostname:
                raise errors.URLError(url)

        # 执行访问
        send_data = {
            "url": url,
        }

        if frame_id:
            send_data["frame_id"] = frame_id

        send_data.update(kwargs)

        return self.__conn.send(
            method="Page.navigate",
            params=send_data,
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def get_frame_tree(
            self,
            raise_err: bool = False,
            need_callback: bool = True,
    ) -> Tuple[bool, dict]:
        """
        获取当前页面的框架树

        https://chromedevtools.github.io/devtools-protocol/tot/Page/#method-getFrameTree

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="Page.getFrameTree",
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def capture_screenshot(
            self,
            format_type: str = "png",
            quality: int = None,
            clip: dict = None,
            from_surface: bool = True,
            capture_beyond_viewport: bool = False,
            optimize_for_speed: bool = False,
            raise_err: bool = False,
            need_callback: bool = True,
    ) -> Tuple[bool, dict]:
        """
        截图

        https://chromedevtools.github.io/devtools-protocol/tot/Page/#method-captureScreenshot

        clip 参考：https://chromedevtools.github.io/devtools-protocol/tot/Page/#type-Viewport

        :param format_type: jpeg, png, webp，默认值为 png
        :param quality: 压缩质量范围为 [0..100]（仅限 jpeg）
        :param clip: 仅捕获给定区域的屏幕截图（{x:1,y:1,width:1,height:1,scale:1}）
        :param from_surface: 从表面捕获屏幕截图，而不是从视图捕获屏幕截图，默认值为 true
        :param capture_beyond_viewport: 捕获视口之外的屏幕截图，默认值为 false
        :param optimize_for_speed: 优化图像编码以提高速度，而不是优化结果大小（默认为 false）
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return: {"data":""}
        """
        if format_type not in ["jpeg", "png", "webp"]:
            format_type = "png"

        send_data = {
            "format": format_type,
            "fromSurface": from_surface,
            "captureBeyondViewport": capture_beyond_viewport,
            "optimizeForSpeed": optimize_for_speed,
        }

        if clip:
            send_data["clip"] = clip
        if quality:
            send_data["quality"] = quality

        return self.__conn.send(
            method="Page.captureScreenshot",
            params=send_data,
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def close(
            self,
            raise_err: bool = False,
            need_callback: bool = False,
    ) -> Tuple[bool, dict]:
        """
        关闭当前页面

        https://chromedevtools.github.io/devtools-protocol/tot/Page/#method-close

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="Page.close",
            raise_err=raise_err,
            need_callback=need_callback,
        )



from typing import Tuple
from easypage.conn import Conn


class Emulation:
    """
        Emulation

        https://chromedevtools.github.io/devtools-protocol/tot/Emulation
    """

    def __init__(self, conn: Conn):
        self.__conn = conn

    def set_focus_emulation_enabled(
            self,
            raise_err: bool = False,
            need_callback: bool = False,
    ) -> Tuple[bool, dict]:
        """
        焦点模拟，可以模拟键盘事件

        https://chromedevtools.github.io/devtools-protocol/tot/Emulation/#method-setFocusEmulationEnabled

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="Emulation.setFocusEmulationEnabled",
            params={
                "enabled": True
            },
            raise_err=raise_err,
            need_callback=need_callback,
        )



from typing import Union, List


class URLError(Exception):
    def __init__(self, url: str):
        super().__init__("该 url 似乎非法：{}".format(url))


class ProxySetError(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class Timeout(Exception):
    def __init__(self, msg: str):
        super().__init__("超时：{}".format(msg))


class ConnectTimeout(Exception):
    def __init__(self, address: str):
        super().__init__("连接超时：{}".format(address))


class BrowserConnectTimeout(Exception):
    def __init__(self, address: str):
        super().__init__("浏览器连接超时：{}".format(address))


class CDPConnClosed(Exception):
    def __init__(self, msg: str):
        super().__init__("连接关闭：{}".format(msg))


# 接收到的消息有错误时触发
class CDPRecvErrorMsg(Exception):
    def __init__(self, msg: str):
        super().__init__("接收到错误消息：{}".format(msg))


class CDPRecvTimeout(Exception):
    def __init__(self, msg: dict):
        super().__init__("接收消息超时：{}".format(msg))


"""
    Element
"""


class ElementWaitTimeout(Exception):
    def __init__(self, ele: Union[List[str], str]):
        super().__init__("元素等待超时：{}".format(ele))


class ElementNotFound(Exception):
    def __init__(self, ele: str):
        super().__init__("没有找到元素：{}".format(ele))



from setuptools import setup, find_packages

setup(
    name='EasyPage',
    version='0.0.7',
    author='leviathangk',
    author_email='1015295213@qq.com',
    description='基于 CDP 协议，简洁快速的浏览器控制 API',
    keywords=['EasyPage', 'page', 'browser'],
    packages=find_packages(),
    install_requires=[
        "lxml",
        "parsel",
        "requests",
        "loguru",
        "beautifulsoup4",
        "websockets",
        "pyee",
        "websocket-client",
    ],
)

"""
    python setup.py sdist
    python -m twine upload dist/*
"""
